//-------------------------------------------------------------------//
//                                                                   //
//  UInfo Copyright (C) Glenn Rosenthal, 1998,1999.  All rights      //
//  reserved.                                                        //
//                                                                   //   
//  The JIS0208.TXT file carries the following copyright             //
//  copyright (c) 1991-1994 Unicode, Inc.  All Rights reserved.      //
//                                                                   //
//  The code page files are also copyright by the Unicode Inc., see  //
//  the individiual files for there copyright noticies.
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//                                                                   //
//  This program was developed to generate UNICODE conversion tables
//  for use with JWPce.  The routine reads a Unicode tranlsation table
//  from the file JIS0208.TXT and converts it into a number of data files 
//  used by JWPce.
//
//  The format of the JIS0208.TXT file is as follows:
//
//      # comment line. (can be many of these)
//      <shift-JIS code><TAB><JIS code><TAB><UNICODE><TAB>#<space><character definition>
//
//  The files produced by the routine are as follows:
//
//      JWP_UKAN.DAT
//
//  This is a C++ array containning UNICODE values and accessed using 
//  JWPce compressed data codes for the index.  This table contains all
//  the coding of the kanji characters to convert them to and from 
//  UNICODE.
//
//  The indexing info this array is as follows:
//
//      index = (<JIS-high_byte>-0x30)*94+<JIS-low_byte>-0x21
//
//      JWP_UMIS.DAT
//
//  This is a simular array containning the UNICODE values for the 
//  symbols and puctuations sets (JIS code pages 0x2100, and 0x2200).
//
//  The indexing info this array is as follows:
//
//      index = (<JIS-high_byte>-0x21)*94+<JIS-low_byte-0x21
//
//-----------------------------------------------------------------------
//
//  In the second version of this code, a number of smaller translation 
//  tables were added.  These tables are specific to translations of the 
//  Windows code pages to UNICODE, and cover only the extnded ASCII 
//  range.  Basically each possible code page uese these 256 characters 
//  differently.  Each of these generates a file of the form: CPxxxx.DAT,
//  that contains 256 numbers that are the mappings for that specific code
//  page.  Each table contains 256 numbers, and is indexed as:
//
//      index = <value>-256
//
//-------------------------------------------------------------------
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "uinfo.h"

//-------------------------------------------------------------------
//
//  Static data.
//
#define NUMBER_KANJI    8000
#define NUMBER_MISC     1000

static int kanji[NUMBER_KANJI];     // Data array for kanji mapping.
static int misc [NUMBER_MISC ];     // Data array for symbol mapping.

//-------------------------------------------------------------------
//
//  Basic parsing routines.
//

//
//  Convert from the code used by JWPce backward to the JIS code.
//
//      code   -- JWPce kanji code (compressed code).
//
//      RETURN -- JIS code.
//
static int code_to_jis (int code) {
  int temp,jis_hi,jis_lo;
  temp   = code/94;
  jis_hi = temp+0x30;
  jis_lo = (code-temp*94)+0x21;
  if ((jis_hi < 0x30) || (jis_hi >= 0x75)) return (0);
  return ((jis_hi << 8) | jis_lo);
}

//
//  Convert from the code used by JWPce backward to extneded ascii code.
//
//      code   -- JWPce kanji code (compressed code).
//
//      RETURN -- JIS code.
//
static int cp_to_jis (int code) {
  if (code >= 0x80) return (0xffff);
  return (code+0x80);
}

//
//  Process a code page (really processes the extended characters of the code page.
//
//      id     -- Code page number
//      text   -- Text description.
//
//      RETURN -- A non-zero value indicates an error.
//
int static do_codepage (int id,char *text) {
  UTable data;
  char   label[100],name[100],output[100];
  sprintf (name  ,"CP%d.TXT",id);
  sprintf (label ,"CodePage-%d",id);
  sprintf (output,"jwp_cp%d.dat",id);
  printf  ("Processing code page %d(%s):\n",id,text);
  if (data.read (name,7,2              )) { printf ("Unable to read code page file\n" ); return (true); }
  if (data.write(output,label,cp_to_jis)) { printf ("Unable to write code page file\n"); return (true); }
  return (false);
}

//
//  Extract a hex value from the first 4 characters of the given buffer.
//
//      buffer -- Buffer to extract data from.
//
//      RETURN -- Extracted hex value.  
//
static int get_hex (char *buffer) {
  int value;
  buffer[4] = 0;
  sscanf (buffer,"%x",&value);
  return (value);
}

//
//  Convert from the code used by JWPce backward to the JIS code.
//
//      code   -- JWPce symbol code (compressed code).
//
//      RETURN -- JIS code.
//
static int misc_to_jis (int code) {
  int temp,jis_hi,jis_lo;
  temp   = code/94;
  switch (temp) {
    case 0: jis_hi = 0x21; break;
    case 1: jis_hi = 0x22; break;
    case 2: jis_hi = 0x28; break;
    default: return (0);
  }
  jis_lo = (code-temp*94)+0x21;
  return ((jis_hi << 8) | jis_lo);
}

//
//  Writes a header block to a file.
//
//      file -- Pointer to open file to write to.
//      type -- String indicating file type ("symbol" or "kanji");
//
void write_head (FILE *file,char *type) {
  fprintf (file,"//-------------------------------------------------------------------//\n");
  fprintf (file,"//                                                                   //\n");
  fprintf (file,"//  JWPce Copyright (C) Glenn Rosenthal, 1998,1999.  All rights      //\n");
  fprintf (file,"//     reserved.                                                     //\n");
  fprintf (file,"//                                                                   //\n");
  fprintf (file,"//  The JIS0208.TXT file carries the following copyright             //\n");
  fprintf (file,"//  copyright (c) 1991-1994 Unicode, Inc.  All Rights reserved.      //\n");
  fprintf (file,"//                                                                   //\n");
  fprintf (file,"//-------------------------------------------------------------------//\n");
  fprintf (file,"\n");
  fprintf (file,"//-------------------------------------------------------------------//\n");
  fprintf (file,"//\n");
  fprintf (file,"// Translate table for %s <-> unicode\n",type);
  fprintf (file,"//\n");
  fprintf (file,"\n");
  return;
}


//-------------------------------------------------------------------
//
//  Begin class UTable
//

//
//  Find a unicode value by index.
//
//      j      -- Index value to search for.
//
//      RETURN -- Unicode value, or a zero if the character cannot be found.
//
int UTable::find (int j) {
  int i;
  for (i = 0; (i < count) && (data[i].index != j); i++);
  if (i == count) return (0);
  return (data[i].unicode);
}


//
//  Read a file into the table.
//
//      name   -- Name of file to read from.
//      u      -- Offset in characters to the start of the unicode column in the table.
//      j      -- Offset in characters to the index colum to be used.
//
//      RETURN -- A non-zero value indicates an error.l
//
int UTable::read (char *name,int u,int j) {
  FILE *file;
  char buffer[500];
  int  i,unicode,index;

  count = 0;
  printf ("Reading unicode data file [%s] ... ",name);
  if (!(file = fopen(name,"r"))) { printf ("\n\nCannot open datafile\n"); return (true); }
  for (i = 0; true; i++) {
    fgets (buffer,sizeof(buffer),file);
    if (feof(file)) break;
    if (buffer[0] == '#') continue;                                 // Skip comments
    if (((int) strlen(buffer) < u) || ((int) strlen(buffer) < j)) continue;     // Skip short lines 
    if (count >= MAX_CHARS) { printf ("\n\nToo many entries in the table\n"); fclose (file); return (true); }
    index   = get_hex(buffer+j);                    // Get jis and unicode values.
    unicode = get_hex(buffer+u);
    if (unicode == 0x301c) unicode = 0xff5e;        // Not actually standard, but machines MS usages (TrueType fonts)
    data[count  ].index   = index;
    data[count++].unicode = unicode;
  }
  printf ("done [%d lines read]\n",i);              // Done!
  fclose (file);
  return (false);
}

//
//  Write the kanji data file.
//
//      name   -- Name of file to write.
//
//      REUTRN -- A non-zero return indicates an error.
//
int UTable::write (char *name,char *label,int (*proc)(int)) {
  FILE *file;
  int   i,j,start;
//
//  Find the highest kanji code value that has a kanji associated with it (this determins
//  the end of the table.
//
  for (j = i = 0; i < 10000; i++) {
    if (find((*proc)(i))) j = i;
  }
//
//  Initialize the file.
//
  printf ("Writing %s unicode table [%s] ... ",label,name);
  if (!(file = fopen(name,"w"))) { printf ("\n\nCannot open database\n"); return (true); }
  write_head (file,label);
  fprintf (file,"  ");                                  // File startup.
//
//  Write the table
//
  for (start = i = 0; i <= j; i++) {
    fprintf (file,"0x%4.4x,",find((*proc)(i)));     // Write entry.
    if (i && !((i+1) % 10)) {                           // Line breaks
      fprintf (file,"  // 0x%4.4x..0x%4.4x\n  ",(*proc)(start),(*proc)(i));
      start = i+1;
    }
  }
  fprintf (file,"  // 0x%4.4x..0x%4.4x\n  ",(*proc)(start),(*proc)(i-1));
//
//  Clean up.
//
  fclose (file);                                        // Finish up.
  printf ("done [%d characters output]\n",j);
  return (false);
}

//
//  End class UTable
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  Main routines.
//

int main (int argc,char **argv) {
  UTable data;
  printf ("UINFO Utility Verison 2.1 copyright (C) 1998,1999,2000 glenn rosenthal, all rights reserved\n\n");
  if (data.read       ("JIS0208.TXT",16,9                 )) { printf ("Unable to read the main UNICODE data file\n"); exit (5); }
  if (data.write      ("jwp_ukan.dat","kanji",code_to_jis )) { printf ("Unable to write kanji trnaslate table\n"); exit (5); }
  if (data.write      ("jwp_umis.dat","symbol",misc_to_jis)) { printf ("Unable to write symbol trnaslate table\n"); exit (5); }
  if (do_codepage(1250,"Eastern Europe"  )) exit (5);
  if (do_codepage(1251,"Cyrillic"        )) exit (5);
  if (do_codepage(1252,"USA, West Europe")) exit (5);
  if (do_codepage(1253,"Greek"           )) exit (5);
  if (do_codepage(1254,"Turkish"         )) exit (5);
  if (do_codepage(1255,"Hebrew"          )) exit (5);
  if (do_codepage(1256,"Arabic"          )) exit (5);
  if (do_codepage(1257,"Baltic"          )) exit (5);
  if (do_codepage(1258,"Vietnamese"      )) exit (5);
  return (0);
}
